/*************************************************************************
  (c) Copyright.  Digital Equipment Corporation, 1995.  All Rights
  Reserved.

  Permission is hereby granted to use, copy, modify, or enhance this 
  software freely, as long as the foregoing copyright of Digital Equipment
  Corporation and this notice are retained on the software.  This 
  software may not be distributed or sublicensed for a fee.  Digital      
  makes this software available "AS IS" and without warranties of any
  kind.  
 *************************************************************************/

/*
* Module Name:
*   cm_drv.h
*   
* Overview:
*   This file describes the interface between the Lower Device Driver and the
*   Connection Manager.  The interface description contains function prototypes
*   as well as details about how the two components must interact.
*
* Authors:
*   TLR - Theodore L. Ross
*
* Modification History:
*   Date        Name  Description 
*   20-Oct-94   TLR   Created.
*   10-Feb-95   TLR   Major modifications made to the interface.
*   17-Feb-95   TLR   Modified based on inputs from team inspection.
*   08-Sep-95   TLR   Added support for buffer pool assignment and QoS.
*
* Description:
*   This header file describes the interface between the Connection Manager
*   (CM) and the device driver.  This interface is used to transfer data
*   packets between the driver and the CM and to set up and tear down ATM
*   virtual channels (VCs).
*
*   This interface describes function calls that go in both directions between
*   the Connection Manager and the Lower Device Driver.  The calls that are
*   implemented in the Lower Device Driver are prefixed with "drv_cm_" which
*   denotes that they are implemented by the Driver and are called by the
*   Connection Manager.  These calls are registered by the device driver prior
*   to their use.  The functions that are implemented in the Connection
*   Manager and called by the Device Driver are prefixed "cm_drv_".
*
* Call Sequence Requirements (Driver implemented functions appear in
* parentheses):
*
*   The cm_drv_call_register function must be called during device driver
*   initialization.  No other calls to functions described in this interface
*   may be made prior to cm_drv_call_register.
*
*   For a specific VC, the following call sequence requirements hold (calls
*   with lesser indentation must precede calls with greater indentation):
*
*     - (drv_cm_vc_setup)
*       - cm_drv_setup_cfm
*         - (drv_cm_xmt)
*           - cm_drv_xmt_done
*         - cm_drv_rcv
*         - (drv_cm_vc_teardown)
*           - cm_drv_teardown_cfm
*
*   Transmit-Done Indications:
*
*       There MUST be a one-to-one correspondence between calls to drv_cm_xmt
*       that return STATUS_K_PENDING and calls to cm_drv_xmt_done.  In other
*       words, for every transmit packet that is designated "pending", there
*       MUST be exactly one call to cm_drv_xmt_done.
*
*       Calls to drv_cm_xmt that result in STATUS_K_SUCCESS status MUST NOT
*       be followed by an associated call to cm_drv_xmt_done.
*
* Example Usage Scenario:
*
*   cm_drv_call_register........During the initialization of the lower device
*                               driver, this call is made into the connection
*                               manager.  It provides the connection manager
*                               with pointers to the driver's routines.
*
*   cm_drv_link_status..........After initialization of the lower device
*                               driver and the ATM interface, the driver
*                               detects the presence of a physical link (i.e.
*                               Sonet synchronization completes, etc.).  This
*                               call is made to notify the connection manager
*                               that the link status is LINK_PHY_UP.
*
*                               The connection manager propagates the link
*                               status to each of its registered SAP clients
*                               some of which may attempt to communicate over
*                               the ATM link.
*
*   (drv_cm_vc_setup)...........This call into the driver is invoked when the
*                               Connection Manager wishes to open a new VC.
*                               The call provides the vpi and vci of the
*                               requested channel plus the quality of service
*                               parameters requested for the VC.
*
*                               It is assumed that the VC setup operation is
*                               asynchronous.  When this call returns, the
*                               VC is not necessarily ready to be used.
*
*   cm_drv_setup_cfm............When the previously requested VC setup has
*                               completed, this call informs the connection
*                               manager that the new VC is ready to be used.
*                               Note that if the VC setup operation is not
*                               asynchronous for a specific ATM interface,
*                               this confirmation call may be invoked by the
*                               driver before the drv_cm_vc_setup routine
*                               completes (i.e. it may be called by the
*                               drv_cm_vc_setup function itself).
*
*   (drv_cm_xmt)................The connection manager uses this call to
*                               transmit a packet to the ATM interface.  If
*                               the packet is copied by the driver it must
*                               return success status indicating that the
*                               transmit buffer may be overwritten or freed.
*
*   (drv_cm_xmt)................Another packet is transmitted by the
*                               connection manager.  In this case, the driver
*                               does not copy the packet (refer to the
*                               function header for the drv_cm_xmt for
*                               reasons why packet should or should not be
*                               copied).  The return status for this call is
*                               "Pending" which indicates to the connection
*                               manager that the transmit buffer must not be
*                               overwritten or freed.
*
*   cm_drv_xmt_done.............Because the lower driver accepted a transmit
*                               packet that it did not copy, the connection
*                               manager must be informed when that
*                               transmit buffer may be overwritten or freed.
*                               This call to the connection manager indicates
*                               that the specified transmit buffer has had its
*                               contents transmitted and may now be reused.
*
*                               This call only occurs for transmit packets for
*                               which "Pending" status was returned.
*
*   cm_drv_rcv..................When a packet is received on the ATM
*                               interface, the device driver uses this call to
*                               pass the packet to the connection manager.
*                               The call passes the address of the receive
*                               packet structure along with the vpi, vci, and
*                               queue for the channel on which the packet was
*                               received.
*
*   (drv_cm_vc_teardown)........The connection manager invokes this call to
*                               request the teardown of a virtual channel.
*                               It is assumed that the teardown process is
*                               ansynchronous.  No packets will be transmitted
*                               on this VC after this call is made.
*
*   cm_drv_teardown_cfm.........This call informs the connection manager that
*                               the previously requested VC teardown is
*                               complete.  The VC is then returned to the pool
*                               of available VCs.  Note that if the VC
*                               teardown operation is not asynchronous, this
*                               call may be invoked by the device driver
*                               before the drv_cm_vc_teardown completes
*                               (i.e. it may be called by the
*                               drv_cm_vc_teardown function itself).
*
*/


/*++
* =====================
* = (DRV_CM_VC_SETUP) =
* =====================
*
* Overview:
*   Call to the driver from the CM for setting up a new VC.  It is assumed
*   that this operation is asynchronous.  If VC setup can be completed
*   quickly and without blocking, this function may call the cm_drv_setup_cfm
*   routine prior to returning.
*
* Arguments:
*   drv_handle       - (IN)  Driver handle associated with this CM instance.
*   vpi              - (IN)  Virtual Path Indicator for this VC.
*   vci              - (IN)  Virtual Channel Indicator for this VC.
*   xmt_pool_context - (IN)  Transmit buffer pool to associate with the VC.
*   rcv_pool_context - (IN)  Receive buffer pool to associate with the VC.
*   p_requested_qos  - (IN)  Requested Quality of Service parameters
*              
* Returns:
*   STATUS_K_PENDING    - Operation will be successfully completed when
*                         cm_drv_setup_cfm is called.
*
* Preconditions:
*   The VC is not currently open.
*
* Postconditions:
*   None
--*/
typedef void (*DRV_CM_VC_SETUP) (HANDLE   drv_handle,
                                 UINT8    vpi,
                                 UINT16   vci,
                                 HANDLE   xmt_pool_context,
                                 HANDLE   rcv_pool_context,
                                 QOS     *p_requested_qos);


/*++
* ========================
* = (DRV_CM_VC_TEARDOWN) =
* ========================
*
* Overview:
*   Call to the driver from the CM for tearing down VCs.  It is assumed that
*   this is an asynchronous call.  If the VC teardown operation can be
*   completed quickly and without blocking, this routine may call the
*   cm_drv_teardown_cfm itself before it returns.
*
*   The function cm_drv_teardown_cfm is called to confirm the completion of
*   the teardown operation.
*
* Arguments:
*   drv_handle         - (IN) Driver handle associated with this CM instance.
*   vpi                - (IN) Virtual Path Indicator for this VC.
*   vci                - (IN) Virtual Channel Indicator for this VC.
*   xmt_pool_context   - (IN) Transmit pool associated with the VC.
*   rcv_pool_context   - (IN) Receive pool associated with the VC.
*              
* Returns:
*   STATUS_K_PENDING - The operation will be complete when the call to
*                      cm_drv_teardown_cfm is made.
*
* Preconditions:
*   The VC is open.
*
* Postconditions:
*   None.
--*/
typedef void (*DRV_CM_VC_TEARDOWN) (HANDLE    drv_handle,
                                    UINT8     vpi,
                                    UINT16    vci,
                                    HANDLE    xmt_pool_context,
                                    HANDLE    rcv_pool_context);


/*++
* ================
* = (DRV_CM_XMT) =
* ================
*
* Overview:
*   This is the function in the driver called by the CM when a packet is to
*   be transmitted.
*
* Arguments:
*   drv_handle       - (IN) Driver handle associated with this CM instance.
*   conn_handle      - (IN) Handle of the VC structure for this connection.  The
*                           Device Driver only needs this parameter if it will call
*                           cm_drv_xmt_done later.  The conn_handle is required
*                           as an argument to cm_drv_xmt_done.
*   p_packet         - (IN) Pointer to the packet to be transmitted.
*   length           - (IN) Length of the packet in bytes.  If this parameter is
*                           zero, the length is unspecified and this module must
*                           find the length by another means.
*   vpi              - (IN) Virtual Path Indicator for the VC.
*   vci              - (IN) Virtual Channel Indicator for the VC.
*   xmt_pool_context - (IN) Transmit pool on which to transmit the packet.
*   user_data        - (IN) User data passed from the SAP to the driver.  Refer to
*                           "system.h" for a complete description of the format of
*                           this argument.
*   p_aal_data       - (IN) AAL specific data that is to be transmitted with the
*                           packet.
*              
* Returns:
*   STATUS_K_SUCCESS    - Packet was transmitted or copied and the buffer may
*                         be freed.
*   STATUS_K_PENDING    - Packet was successfully queued for transmit.  The
*                         buffer may not yet be freed.
*   STATUS_K_CONGESTED  - Packet was not queued for transmit and was not
*                         discarded.
*   STATUS_K_FAILURE    - There was a failure in the transmission of the
*                         packet.  The packet was not transmitted and must be
*                         discarded.
*
* Preconditions:
*   The vpi/vci is correctly mapped to the xmt_queue and has been opened.
*
* Postconditions:
*   None.
*
* Behavior Requirements:
*   If the "internal_source" flag in the user_data is set, the entire transmit
*   packet MUST be copied into another transmit buffer.  The return status
*   MUST NOT be STATUS_K_PENDING.
*
*   If the "internal_source" flag in the user_data is set, the length argument
*   is assumed to be non-zero.  If it is zero, no data may be transmitted
*   and the return status from the call MUST be STATUS_K_FAILURE.
*
*   If the "add_le_header" flag in the user_data is set, the 16-bit "le_header"
*   field MUST be prepended to the transmit packet before it is transmitted to
*   the network.
--*/
typedef STATUS (*DRV_CM_XMT) (HANDLE    drv_handle,
                              HANDLE    conn_handle,
                              void     *p_packet,
                              UINT32    length,
                              UINT8     vpi,
                              UINT16    vci,
                              HANDLE    xmt_pool_context,
                              UINT32    user_data,
                              AAL_DATA *p_aal_data);


/*++
* =================
* = cm_drv_config =
* =================
*
* Overview:
*   Provides the CM instance with the lower device driver handle.
*
* Arguments:
*   cm_handle   - (IN)  Handle returned by cm_create.
*   drv_handle  - (IN)  Handle of the Lower Device Driver instance.
*              
* Returns:
*   None
*
* Preconditions:
*   The cm_handle must be valid.
*
* Postconditions:
*   None
--*/
void cm_drv_config (HANDLE  cm_handle,
                    HANDLE  drv_handle);

/*++
* ========================
* = cm_drv_call_register =
* ========================
*
* Overview:
*   Register the driver's routines with the Connection Manager.  This call
*   should be made as early as possible in the initialization of the ATM
*   subsystem.
*
*   The interface between the CM and the Lower Driver will not operate until
*   the calls are registered.
*
* Arguments:
*   cm_handle            - (IN) Handle of the associated CM instance.
*   drv_cm_vc_setup      - (IN) Address of the VC Setup Call.
*   drv_cm_vc_teardown   - (IN) Address of the VC Teardown Call.
*   drv_cm_xmt           - (IN) Address of the Transmit Call.
*              
* Returns:
*   None
*
* Preconditions:
*   None
*
* Postconditions:
*   None
--*/
void cm_drv_call_register(HANDLE                cm_handle,
                          DRV_CM_VC_SETUP       drv_cm_vc_setup,
                          DRV_CM_VC_TEARDOWN    drv_cm_vc_teardown,
                          DRV_CM_XMT            drv_cm_xmt);

/*****************************************************************************
 * Buffer Pool Handling Interface
 *
 *   The following definitions and functions are used by the lower device
 *   driver to manage transmit and receive buffer pools.
 *
 ****************************************************************************/

typedef enum
   {
   POOL_XMT,
   POOL_RCV
   } POOL_TYPE;

#define POOL_MASK_ANY_VC      0x00000001
#define POOL_MASK_CBR_VCS     0x00000002
#define POOL_MASK_VBR_VCS     0x00000004
#define POOL_MASK_UBR_VCS     0x00000008
#define POOL_MASK_ABR_VCS     0x00000010
#define POOL_MASK_FM_VCS      0x00000020
#define POOL_MASK_QFC_VCS     0x00000040

/*++
* ========================
* = cm_drv_pool_register =
* ========================
*
* Overview:
*   This call is used by the lower device driver to register buffer pools with
*   the Connection Manager.  At least one transmit and one receive buffer pool
*   must be registered before the CM can open VCs.
*
* Arguments:
*   cm_handle            - (IN)  Handle of the associated CM instance.
*   pool_context         - (IN)  Context to be used in calls to the lower driver.
*                                This is simply stored by the CM, never read or
*                                modified.
*   pool_type            - (IN)  Direction of flow through the pool
*                                (POOL_XMT or POOL_RCV).
*   buffer_size_in_bytes - (IN)  Size of the buffers in the buffer pool.
*   depth_in_buffers     - (IN)  Number of buffers in the buffer pool.
*   assignment_mask      - (IN)  Bit mask containing VC assignment criteria.
*   p_pool_handle        - (OUT) Returned handle for the registered pool.
*              
* Returns:
*   STATUS_K_SUCCESS    - Successful pool registration.
*   STATUS_K_RESOURCES  - Insufficient resources to complete registration.
*
* Preconditions:
*   None
*
* Postconditions:
*   None
--*/
STATUS cm_drv_pool_register (HANDLE         cm_handle,
                             HANDLE         pool_context,
                             POOL_TYPE      pool_type,
                             UINT32         buffer_size_in_bytes,
                             UINT32         depth_in_buffers,
                             UINT32         assignment_mask,
                             HANDLE        *p_pool_handle);

/*++
* ==========================
* = cm_drv_pool_deregister =
* ==========================
*
* Overview:
*   This call is used by the lower device driver to remove a buffer pool
*   that was previously registered.
*
* Arguments:
*   cm_handle     - (IN) Handle of the associated CM instance.
*   pool_handle   - (IN) Handle returned by cm_drv_pool_register.
*              
* Returns:
*   STATUS_K_SUCCESS    - Successful pool deregistration.
*
* Preconditions:
*   None
*
* Postconditions:
*   The deregistered pool is no longer available for assignment by the CM and
*   any VCs associated with the buffer pool have been closed.
--*/
STATUS cm_drv_pool_deregister (HANDLE  cm_handle,
                               HANDLE  pool_handle);


/*++
* ====================
* = cm_drv_setup_cfm =
* ====================
*
* Overview:
*   Call from the lower driver to the CM that confirms that a previously
*   requested VC setup has completed.
*
* Arguments:
*   cm_handle     - (IN) CM handle associated with this driver instance.
*   status        - (IN) Completion status of the setup operation. May be:
*                         STATUS_K_SUCCESS - Operation completed successfully,
*                         STATUS_K_FAILURE - Operation failed.
*   vpi           - (IN) Virtual Path Indicator for the VC.
*   vci           - (IN) Virtual Channel Indicator for the VC.
*   p_actual_qos  - (IN) Actual QoS of VC.
*              
* Returns:
*   None
*
* Preconditions:
*   None
*
* Postconditions:
*   The mapping has been updated to contain this VC.
--*/
void cm_drv_setup_cfm (HANDLE  cm_handle,
                       STATUS  status,
                       UINT8   vpi,
                       UINT16  vci,
                       QOS    *p_actual_qos);

/*++
* =======================
* = cm_drv_teardown_cfm =
* =======================
*
* Overview:
*   Call from the lower driver to the CM that confirms that a previously
*   requested VC teardown has completed.
*
* Arguments:
*   cm_handle   - (IN) CM handle associated with this driver instance.
*   vpi         - (IN) Virtual Path Indicator for the VC.
*   vci         - (IN) Virtual Channel Indicator for the VC.
*              
* Returns:
*   None
*
* Preconditions:
*   None
*
* Postconditions:
*   There is no VC entry in the mapping for this vpi/vci.
--*/
void cm_drv_teardown_cfm (HANDLE  cm_handle,
                          UINT8   vpi,
                          UINT16  vci);

/*++
* ==============
* = cm_drv_rcv =
* ==============
*
* Overview:
*   This is the function in the CM called by the lower driver when a packet has
*   been received.
*
* Arguments:
*   cm_handle   - (IN) CM handle associated with this driver instance.
*   length      - (IN) Length of the received packet in bytes.  Since the
*                      device driver is an AAL driver, it is assumed that
*                      this length is the length of the packet after it has
*                      had any AAL encapsulation (i.e. AAL5 trailer) removed.
*   vpi         - (IN) Virtual Path Indicator for the VC.
*   vci         - (IN) Virtual Channel Indicator for the VC.
*   user_data   - (IN) User data to be passed from the driver to the SAP.
*   p_aal_data  - (IN) AAL specific data received with the packet.
*   pp_packet   - (IO) Pointer to a buffer.  For input, this is a pointer to
*                      the received packet.  On output (after the call
*                      returns) this a pointer to a free buffer that is
*                      available for re-use.  If this argument is NULL on
*                      return, there was no free buffer available to be
*                      returned.
*              
* Returns:
*   None.
*
* Preconditions:
*   None.
*
* Postconditions:
*   The returned packet pointer is either NULL or points to a buffer that may
*   be overwritten or otherwise destroyed. The upper layers may pass the
*   received packet on and return a pointer to a fresh buffer from a free pool
*   or may copy (or discard) the packet and return the same pointer that was
*   provided in the original call.
--*/
void cm_drv_rcv (HANDLE      cm_handle,
                 UINT32      length,
                 UINT8       vpi,
                 UINT16      vci,
                 UINT32      user_data,
                 AAL_DATA   *p_aal_data,
                 void      **pp_packet);

/*++
* ===================
* = cm_drv_xmt_done =
* ===================
*
* Overview:
*   Notifies the CM that a pending transmit operation has completed.  This
*   call should be made when the buffer containing a transmit packet has been
*   completely DMA'd across the bus.  It is used by the upper layers to free
*   the buffer for re-use.
*
*   This call MUST be made ONLY for transmit buffers that were passed by
*   calls to drv_cm_xmt that resulted in STATUS_K_PENDING status.  This call
*   MUST be made EXACTLY once for each such transmit buffer.
*
* Arguments:
*   cm_handle   - (IN) CM handle associated with this driver instance.
*   conn_handle - (IN) Connection handle passed with the transmit request.
*   p_packet    - (IN) Pointer to the packet buffer that was transmitted.
*   user_data   - (IN) User data that was supplied with the transmit packet.
*              
* Returns:
*   None.
*
* Preconditions:
*   None.
*
* Postconditions:
*   None.
--*/
void cm_drv_xmt_done (HANDLE    cm_handle,
                      HANDLE    conn_handle,
                      void     *p_packet,
                      UINT32    user_data);

