/*************************************************************************
  (c) Copyright.  Digital Equipment Corporation, 1995.  All Rights
  Reserved.

  Permission is hereby granted to use, copy, modify, or enhance this 
  software freely, as long as the foregoing copyright of Digital Equipment
  Corporation and this notice are retained on the software.  This 
  software may not be distributed or sublicensed for a fee.  Digital      
  makes this software available "AS IS" and without warranties of any
  kind.  
 *************************************************************************/

/*
* Module Name:
*   lec_arp.c
*   
* Overview:
*   This file implements the LAN Emulation ARP Cache.
*
* Authors:
*   TLR - Theodore L. Ross
*
* Modification History:
*   Date       Name  Description 
*   29-Sep-94  TLR   Created.
*   22-Mar-95  TLR   First Implementation Completed.
*
* Suspicions:
*
*/

#ifdef _DEBUG
#define STATIC
#include <stdlib.h>
#include <stdio.h>
#include <assert.h>
#else
#define STATIC static
#define assert(expr)
#endif

/* Include General Headers. */

#include "g_types.h"    /* General Purpose Types      */
#include "codes.h"      /* Return Status Codes        */
#include "atm.h"        /* ATM Specific Definitions   */
#include "utl.h"        /* General Purpose Utilities  */
#include "utl_os.h"     /* Operating System Utilities */
#include "system.h"
#include "g_event.h"

/* Include Headers for Mapping Module Interfaces. */

#include "af_lane.h"
#include "cm.h"
#include "map.h"
#include "svc_info.h"
#include "map_data.h"
#include "lec_mgt.h"
#include "lec.h"
#include "lec_arp.h"

#ifdef PROXY
#include "proxy.h"
#endif

#define EMOD   MOD_LEC_ARP
#define EINST  p_elan->p_text
#define EMASK  p_elan->event_mask


/* Define the LE-ARP refresh interval.  This is the number of milliseconds
 * that elapse between LE-ARP requests sent to refresh the LE-ARP cache.
 * Note that if this value is large, there will be infrequent servicing of
 * the arp queue.  This will decrease the CPU overhead but LE-ARP requests
 * will be batched in relatively large groups causing broadcast bursts.
 * If the value is low, the LE-ARPs are sent in a smoother fashion but at the
 * cost of slightly higher CPU overhead.
 */
#define REFRESH_INTERVAL      500

#define USE_AGE_LIMIT         6000000

static void esi_linkage_remove (HANDLE esi_handle);

/* EVENT_QUEUE
 *
 *  Descriptor for an event queue.
 *
 *  total_time  - Total delay represented by the elements in the event queue.
 *                This is the time in milliseconds from the present moment
 *                that the last event in the queue will occur.
 *
 *                The total_time value is used as an optimization for adding
 *                events to the end of the queue.  It allows events to be
 *                added to the end of the queue without the need for traversing
 *                the entire queue.
 */
typedef struct event_tag
   {
   struct age_ref    *p_head;
   struct age_ref    *p_tail;
   UINT32             size;
   UINT32             total_time;
   } EVENT_QUEUE;

typedef struct _elan
   {
   struct _elan  *p_next;
   char                *p_text;
   UINT32               event_mask;
   HANDLE               la_handle;
   HANDLE               lc_elan_handle;
   HANDLE               lport_handle;
   HANDLE               esi_list;
   BOOLEAN              configured;
   EVENT_QUEUE          fixed_event_queue;
   EVENT_QUEUE          variable_event_queue;
   LAN_TYPE             lan_type;
   BOOLEAN              proxy_flag;
   BOOLEAN              topology_change_flag;
   UINT16               max_unknown_frame_count;
   UINT16               max_unknown_frame_time;
   UINT32               vcc_timeout;
   UINT16               max_retry_count;
   UINT32               aging_time;
   UINT16               forward_delay_time;
   UINT16               le_arp_response_time;
   UINT16               flush_timeout;
   UINT16               path_switching_delay;
   UINT32               cache_size;
   UINT32               cache_index;
   } LA_ELAN_CONTEXT;

SIMPLE_LIST (LA_ELAN_CONTEXT, LA_ELAN_LIST);

typedef struct
   {
   HANDLE               os_handle;
   HANDLE               lc_handle;
   HANDLE               cm_handle;
   HANDLE               vc_list;
   HANDLE               atm_list;
   ARP_XMT_CALLBACK     arp_xmt_callback;
   FLUSH_XMT_CALLBACK   flush_xmt_callback;
   SVC_SETUP_CALLBACK   svc_setup_callback;
   HANDLE               one_sec_clock;
   HANDLE               arp_refresh_clock;
   UINT32               time_since_creation;
   LA_ELAN_LIST         elan_list;
   } LA_CONTEXT;

STATIC void age_ref_unlink (AGE_REF *p_ref);

/*++
* ========================
* = link_update_callback =
* ========================
*
* Status:
*
* Overview:
*
* Arguments:
*              
* Returns:
*
* Preconditions:
*
* Postconditions:
*
* Description:
*
* Issues:
*
--*/
static void link_update_callback (HANDLE la_handle,
                                  HANDLE esi_handle)
   {
   LA_CONTEXT        *p_context;
   LA_ELAN_CONTEXT   *p_elan;
   MAP_ESI_RECORD    *p_esi;
   MAP_ATM_RECORD    *p_atm;
#ifdef PROXY
   UINT8              vpi;
   UINT16             vci;
#endif

   p_esi     = (MAP_ESI_RECORD  *) esi_handle;
   p_atm     = (MAP_ATM_RECORD  *) p_esi->atm_link;
   p_elan    = (LA_ELAN_CONTEXT *) p_esi->la_elan_handle;
   p_context = (LA_CONTEXT      *) p_elan->la_handle;

   assert (p_atm != NULL);

   p_esi->vc_link = p_atm->active_vc_link;

   switch (p_atm->atm_status)
      {
      case ATM_NONE:

         /* If there is no VC associated with the ATM address and signaling
          * is not in progress.  Set the ARP cache entry to "UNKNOWN".  It
          * will age normally if unused.  If a packet is transmitted to this
          * ESI, the LE-ARP process will begin at that time to reestablish a
          * data-direct VCC.  Note that esi_linkage_remove sets the esi_status
          * to "UNKNOWN".
          */
         esi_linkage_remove (esi_handle);
         break;

      case ATM_PENDING:

         p_esi->esi_status = ESI_VC_PENDING;
         break;

      case ATM_VALID:

         /* If the esi status is VC_PENDING, initiate the flush protocol for
          * this esi.  Otherwise, leave the esi status alone.
          */
         if (p_esi->esi_status == ESI_VC_PENDING)
            {
            if (p_elan->max_unknown_frame_count == 0)
          {
               p_esi->esi_status = ESI_FORWARD_DIRECT;
#ifdef PROXY
               conn_vcid_get ((HANDLE) p_esi->vc_link, &vpi, &vci);
               proxy_data_direct_add (p_elan->lport_handle,
                                      p_esi->esi,
                                      vpi,
                                      vci,
                                      PROXY_FORMAT_802_3,  /* Temporary */
                                      (HANDLE) p_esi->vc_link);
#endif
          }
            else
               {
               p_esi->esi_status = ESI_FLUSH_PENDING;
               p_context->flush_xmt_callback (p_elan->lc_elan_handle,
                                              p_atm->addr,
                                              p_esi);
               }
            }
         break;

      case ATM_FAIL:

         p_esi->esi_status = ESI_UNKNOWN;
         break;

      default:
         assert (FALSE);
      }
   }

/*++
* ========================
* = link_remove_callback =
* ========================
*
* Status:
*
* Overview:
*
* Arguments:
*              
* Returns:
*
* Preconditions:
*
* Postconditions:
*
* Description:
*
* Issues:
*
--*/
static void link_remove_callback (HANDLE la_handle,
                                  HANDLE esi_handle)
   {
   /* Not Implemented */
   }

/*++
* ======================
* = esi_linkage_remove =
* ======================
*
* Status:
*
* Overview:
*
* Arguments:
*              
* Returns:
*
* Preconditions:
*
* Postconditions:
*
* Description:
*
* Issues:
*
--*/
static void esi_linkage_remove (HANDLE esi_handle)
   {
   LA_CONTEXT        *p_context;
   LA_ELAN_CONTEXT   *p_elan;
   MAP_ESI_RECORD    *p_esi;
   MAP_ATM_RECORD    *p_atm;
   HANDLE_RECORD     *p_handle;

   /* Get a pointer to the ESI record structure.  Set the esi_status to
    * ESI_UNKNOWN.  This is done so that if the ESI is not deleted, the next
    * packet transmitted to this ESI causes an LE-ARP to be initiated.
    */
   p_esi = (MAP_ESI_RECORD *) esi_handle;
   p_esi->esi_status = ESI_UNKNOWN;

   /* If the atm_link is NULL, there is no further work to do. */

   if (p_esi->atm_link == NULL)
      return;

   /* Obtain the remaining structure references needed. */

   p_atm     = (MAP_ATM_RECORD  *) p_esi->atm_link;
   p_elan    = (LA_ELAN_CONTEXT *) p_esi->la_elan_handle;
   p_context = (LA_CONTEXT      *) p_elan->la_handle;

   /* Nullify the link from the ESI record to the ATM record. */

   p_esi->atm_link = NULL;

   /* Search the list of back links from the ATM record to ESI records.
    * Delete and deallocate the link that references this ESI record.
    */
   utl_list_traverse (p_atm->esi_link_list, p_handle)
      {
      if (p_handle->handle == (HANDLE) p_esi)
         {
         utl_list_delete (p_atm->esi_link_list, p_handle);
         os_mem_dealloc (p_context->os_handle, p_handle);
         break;
         }
      }

   /* Unlink the esi record's age reference from whichever event queue it
    * is in.
    */
   age_ref_unlink (p_esi->p_age_ref);
   }

/*++
* ================
* = esi_atm_link =
* ================
*
* Status:
*
* Overview:
*
* Arguments:
*              
* Returns:
*
* Preconditions:
*
* Postconditions:
*
* Description:
*
* Issues:
*
--*/
static STATUS esi_atm_link (HANDLE     la_elan_handle,
                            HANDLE     esi_handle,
                            ADDR_ATM   addr_atm,
                            HANDLE    *p_atm_handle)
   {
   LA_CONTEXT        *p_context;
   LA_ELAN_CONTEXT   *p_elan;
   MAP_ESI_RECORD    *p_esi;
   MAP_ATM_RECORD    *p_atm;
   HANDLE             atm_handle;
   STATUS             status;
   HANDLE_RECORD     *p_handle;
   BOOLEAN            created;

   /* Obtain the needed structure references. */

   p_elan    = (LA_ELAN_CONTEXT *) la_elan_handle;
   p_context = (LA_CONTEXT      *) p_elan->la_handle;
   p_esi     = (MAP_ESI_RECORD  *) esi_handle;

   /* Check to see if the ESI record already has linkage to an ATM record.
    * If it does, look at the ATM record to determine if its address matches
    * addr_atm.  If the linked ATM record does not contain the correct ATM
    * address, it must be unlinked from the ESI record so a new one can take
    * its place.
    */
   p_atm = NULL;
   if (p_esi->atm_link != NULL)
      {
      p_atm = (MAP_ATM_RECORD *) p_esi->atm_link;

      if (!ATM_EQUAL (addr_atm, p_atm->addr))
         {
         esi_linkage_remove ((HANDLE) p_esi);
         p_atm = NULL;
         }
      }

   /* At this point, if p_atm is non-NULL, it references an ATM record that
    * contains the correct ATM address and is linked with the ESI record.
    * If p_atm is NULL, we must search for a matching entry in the atm_list.
    * If none is found, create a new one.
    */
   created = FALSE;
   if (p_atm == NULL)
      {
      status = map_key_search (p_context->atm_list, &addr_atm, &atm_handle);
      if (status != STATUS_K_SUCCESS)
         {
         /* Create a new ATM record and populate its fields. */

         p_atm = (MAP_ATM_RECORD *) os_mem_alloc (p_context->os_handle,
                                                  sizeof (MAP_ATM_RECORD));
         if (p_atm == NULL)
            return STATUS_K_RESOURCES;

         ATM_COPY (addr_atm, p_atm->addr);
         p_atm->atm_status     = ATM_NONE;
         p_atm->active_vc_link = NULL;
         utl_list_init (p_atm->vc_link_list);
         utl_list_init (p_atm->esi_link_list);

         /* Insert the new record into the ATM list. */

         map_element_insert (p_context->atm_list, p_atm);
         created = TRUE;
         }
      else
         {
         p_atm = (MAP_ATM_RECORD *) atm_handle;
         }

      /* Link the ESI record and the ATM record. */

      p_esi->atm_link = (HANDLE) p_atm;
      p_esi->vc_link  = p_atm->active_vc_link;

      /* Create a handle record to contain the back link from the ATM record
       * to the ESI record.
       */
      p_handle = (HANDLE_RECORD *) os_mem_alloc (p_context->os_handle,
                                                 sizeof (HANDLE_RECORD));
      if (p_handle == NULL)
         {
         if (created)
            {
            map_element_delete (p_context->atm_list, p_atm);
            os_mem_dealloc (p_context->os_handle, p_atm);
            }
         return STATUS_K_RESOURCES;
         }

      p_handle->handle = (HANDLE) p_esi;
      utl_list_add (p_atm->esi_link_list, p_handle);
      }

   *p_atm_handle = (HANDLE) p_atm;
   return STATUS_K_SUCCESS;
   }

/*++
* ===================
* = esi_record_make =
* ===================
*
* Status:
*
* Overview:
*
* Arguments:
*              
* Returns:
*
* Preconditions:
*
* Postconditions:
*
* Description:
*
* Issues:
*
--*/
static STATUS esi_record_make (HANDLE     la_elan_handle,
                               ESI        esi,
                               BOOLEAN    atm_present,
                               ADDR_ATM   addr_atm,
                               HANDLE    *p_esi_handle,
                               HANDLE    *p_atm_handle)
   {
   LA_CONTEXT        *p_context;
   LA_ELAN_CONTEXT   *p_elan;
   MAP_ESI_RECORD    *p_esi;
   STATUS             status;

   /* Obtain the needed structure references. */

   p_elan    = (LA_ELAN_CONTEXT *) la_elan_handle;
   p_context = (LA_CONTEXT      *) p_elan->la_handle;

   /* Allocate a new ESI record structure. */

   p_esi = (MAP_ESI_RECORD *) os_mem_alloc (p_context->os_handle,
                                            sizeof (MAP_ESI_RECORD));
   if (p_esi == NULL)
      return STATUS_K_RESOURCES;

   /* Allocate an age reference for the ESI record. */

   p_esi->p_age_ref = (AGE_REF *) os_mem_alloc (p_context->os_handle,
                                                sizeof (AGE_REF));
   if (p_esi->p_age_ref == NULL)
      {
      os_mem_dealloc (p_context->os_handle, p_esi);
      return STATUS_K_RESOURCES;
      }

   /* Populate the fields of the new ESI record. */

   ESI_COPY (esi, p_esi->esi);
   p_esi->esi_status                = ESI_UNKNOWN;
   p_esi->atm_link                  = NULL;
   p_esi->vc_link                   = NULL;
   p_esi->la_elan_handle            = la_elan_handle;
   p_esi->remote_flag               = TRUE;
   p_esi->permanent_entry           = FALSE;
   p_esi->arp_age                   = 0;
   p_esi->use_age                   = 0;
   p_esi->tick_count                = 0;
   p_esi->packets_flooded           = 0;
   p_esi->flush_tran_id             = 0;

   /* Populate the fields of the new age reference. */

   p_esi->p_age_ref->esi_handle     = (HANDLE) p_esi;
   p_esi->p_age_ref->p_event_queue  = NULL;
   p_esi->p_age_ref->delta_time     = 0;
   p_esi->p_age_ref->arp_retries    = 0;

   /* If an ATM address was supplied, link the ATM address to the new ESI
    * record.
    */
   if (atm_present)
      {
      status = esi_atm_link (la_elan_handle,
                             (HANDLE) p_esi,
                             addr_atm,
                             p_atm_handle);

      if (status != STATUS_K_SUCCESS)
         {
         os_mem_dealloc (p_context->os_handle, p_esi->p_age_ref);
         os_mem_dealloc (p_context->os_handle, p_esi);
         return STATUS_K_RESOURCES;
         }
      }

   /* Add the new ESI record to the esi_list. */

   map_element_insert (p_elan->esi_list, (HANDLE) p_esi);

   *p_esi_handle = (HANDLE) p_esi;
   return STATUS_K_SUCCESS;
   }

/****************************************************************************
 *  Event Queue Functions
 ****************************************************************************/

/*++
* ====================
* = event_queue_init =
* ====================
*
* Status:
*
* Overview:
*
* Arguments:
*              
* Returns:
*
* Preconditions:
*
* Postconditions:
*
* Description:
*
* Issues:
*
--*/
static void event_queue_init (EVENT_QUEUE *p_event_queue)
   {
   p_event_queue->p_head      = NULL;
   p_event_queue->p_tail      = NULL;
   p_event_queue->size        = 0;
   p_event_queue->total_time  = 0;
   }

/*++
* ======================
* = event_queue_insert =
* ======================
*
* Overview:
*   Schedule an event by placing it in the proper location on an event queue.
*
* Arguments:
*   p_queue         (IN) Pointer to the event queue to be modified.
*   p_age_ref       (IN) Pointer to the age reference to add to the event
*                        queue.
*   absolute_time   (IN) The time in milliseconds from now when the event
*                        is to be scheduled.
*              
* Returns:
*   None
*
* Preconditions:
*   The age reference is not a member of any event queue.
*
* Postconditions:
*   The age reference is a member of the selected event queue and is scheduled
*   to be serviced absolute_time milliseconds from the current time.
*
* Description:
*   If the absolute time is greater than or equal to the event queue's total
*   time, place the new event on the tail of the queue and use the total time
*   to calculate its delat time.
*
*   If the absolute time is less than the queue's total time, traverse the
*   event queue to find the appropriate insertion point.
*
* Issues:
*
--*/
STATIC void event_queue_insert (EVENT_QUEUE *p_queue,
                                AGE_REF     *p_age_ref,
                                UINT32       absolute_time)
   {
   AGE_REF  *p_ptr;
   UINT32    accum_time;

   /* Place the pointer to the event queue in the age reference.  This will
    * be used later in case this reference is prematurely removed from the
    * event queue.  The pointer is used to determine which event queue the
    * reference should be removed from.
    */
   assert (p_age_ref->p_event_queue == NULL);
   p_age_ref->p_event_queue = p_queue;

   /* Check to see if the absolute time is greater than or equal to the event
    * queue's total time.
    */
   if (absolute_time >= p_queue->total_time)
      {
      /* Handle the optimal case.  No traversal is needed, we can insert the
       * new reference at the tail of the event queue.
       *
       * Calculate the reference's delta time based on the queue's total time
       * and adjust the total time to include the new reference.
       */
      p_age_ref->delta_time = absolute_time - p_queue->total_time;
      p_queue->total_time   = absolute_time;

      /* Link the reference to the tail of the event queue. */

      p_age_ref->p_prev     = p_queue->p_tail;
      p_age_ref->p_next     = NULL;
      p_queue->p_tail       = p_age_ref;
      p_queue->size        += 1;

      /* Check to see if this is the only reference in the queue.  If so,
       * point the head pointer to the reference.  If not, resolve the linkage
       * to the previous reference.
       */
      if (p_age_ref->p_prev != NULL)
         {
         p_age_ref->p_prev->p_next = p_age_ref;
         }
      else
         {
         p_queue->p_head = p_age_ref;
         assert (p_queue->size == 1);
         }
      }
   else
      {
      /* Handle the non-optimal case.  The age queue must be traversed to find
       * the proper insertion point.
       */
      p_ptr      = p_queue->p_head;
      accum_time = p_ptr->delta_time;
      while (absolute_time >= accum_time)
         {
         p_ptr       = p_ptr->p_next;

         /* The following assertion can be made because we know that if the
          * end of the queue is reached, the optimal case above would have
          * been selected.
          */
         assert (p_ptr != NULL);
         accum_time += p_ptr->delta_time;
         }

      /* The proper position has been found, take one step back to set up for
       * the insertion.
       */
      accum_time -= p_ptr->delta_time;
      p_ptr       = p_ptr->p_prev;

      /* Calculate the reference's delta time. */

      p_age_ref->delta_time = absolute_time - accum_time;

      /* Insert the new element after p_ptr. */

      if (p_ptr == NULL)
         {
         p_age_ref->p_prev         = NULL;
         p_age_ref->p_next         = p_queue->p_head;
         p_queue->p_head           = p_age_ref;
         p_age_ref->p_next->p_prev = p_age_ref;
         }
      else
         {
         p_age_ref->p_prev         = p_ptr;
         p_age_ref->p_next         = p_ptr->p_next;
         p_ptr->p_next             = p_age_ref;
         p_age_ref->p_next->p_prev = p_age_ref;
         }

      p_queue->size += 1;

      /* Adjust the delta time of the reference that follows the inserted
       * reference.  This ensures that the absolute times of the following
       * references are not affected by this insertion.
       */
      assert (p_age_ref->p_next->delta_time > p_age_ref->delta_time);
      p_age_ref->p_next->delta_time -= p_age_ref->delta_time;
      }
   }

/*++
* ======================
* = event_queue_remove =
* ======================
*
* Overview:
*   Service an event queue.  If there is an event scheduled to occur during
*   this interval, remove its reference from the queue and return it to the
*   caller.
*
* Arguments:
*   p_queue         - (IO)  Pointer to the queue to be serviced.
*
*   p_interval_time - (IO)  Pointer to the current interval time in msec.
*                           This value is decremented by the time consumed
*                           by this call.
*
*   pp_age_ref      - (OUT) Pointer to a pointer to be returned.  The returned
*                           pointer is to the reference to be serviced.  If
*                           no events are scheduled for this interval, the
*                           returned pointer is NULL.
*              
* Returns:
*   None
*
* Preconditions:
*   None
*
* Postconditions:
*   The total time represented by the event queue is decreased by the
*   difference between the input and output values of the interval time.
*
*   If the head element was scheduled to be serviced in this time interval,
*   the interval time is decreased by the head reference's delta time and the
*   pointer to the reference is returned.
*
*   If the head element was not scheduled to be serviced in this interval,
*   the interval time is decreased to zero and the returned pointer is NULL.
*
* Recommended Usage:
*   It is recommended that this routine be used in a loop.  This is because
*   several events may be scheduled for the same time interval.  The loop
*   should continue until the returned pointer is NULL.  Refer to the following
*   code example:
*
*       map_event_queue_remove (&queue, &int_time, &p_ref);
*       while (p_ref != NULL)
*          {
*          service_event (p_ref);
*          map_event_queue_remove (&queue, &int_time, &p_ref);
*          }
*
* Description:
*
* Issues:
*
--*/
STATIC void event_queue_remove (EVENT_QUEUE *p_queue,
                                UINT32      *p_interval_time,
                                AGE_REF    **pp_age_ref)
   {
   /* Handle the case where there are no entries on the event queue.  In this
    * case, the interval time is not relevant.  Return a NULL age reference
    * pointer.
    */
   if (p_queue->p_head == NULL)
      {
      *pp_age_ref = NULL;
      return;
      }

   /* Handle the case in which the head element has a delta time that is
    * greater than the interval time.  In this case, both the total time and
    * the head element's delta time are decreased by the interval time.  Since
    * there are no events which are due to occur during this interval, return
    * a null age reference pointer.
    *
    * Set the interval time to zero.  This indicates that the entire interval
    * has been consumed.
    */
   if (p_queue->p_head->delta_time > *p_interval_time)
      {
      p_queue->p_head->delta_time -= *p_interval_time;
      p_queue->total_time         -= *p_interval_time;
      *p_interval_time             = 0;
      *pp_age_ref                  = NULL;
      return;
      }

   /* The final case is the one in which the head element in the queue has
    * a delta time less than or equal to the interval time.  In this case, the
    * head element is scheduled to be serviced during this interval.  The
    * interval time is decreased by the head element's delta time to indicate
    * how much of the time interval is left for the next element.
    *
    * Dequeue the head element and return a pointer to it.
    */
   *p_interval_time     -= p_queue->p_head->delta_time;
   p_queue->total_time  -= p_queue->p_head->delta_time;
   *pp_age_ref           = p_queue->p_head;
   p_queue->p_head       = p_queue->p_head->p_next;
   p_queue->size        -= 1;
   if (p_queue->p_head == NULL)
      {
      p_queue->p_tail = NULL;
      assert (p_queue->size == 0);
      }
   else
      {
      p_queue->p_head->p_prev = NULL;
      }

   /* Since this event is no longer on an event queue (having just been
    * removed), nullify the event queue reference.
    */
   (*pp_age_ref)->p_event_queue = NULL;
   }


/*++
* ==================
* = age_ref_unlink =
* ==================
*
* Overview:
*   Remove a reference from an event queue.  This is used to unschedule an
*   event.
*
* Arguments:
*   p_ref   - (IN)  Pointer to the reference to be removed.
*              
* Returns:
*   None
*
* Preconditions:
*   None
*
* Postconditions:
*   The reference is not scheduled in an event queue.
*
* Description:
*
* Issues:
*
--*/
STATIC void age_ref_unlink (AGE_REF *p_ref)
   {
   /* If the pointer to the age reference is null, no action is necessary. */

   if (p_ref != NULL)
      {
      /* If this age reference is not a member of an event queue, no action
       * is necessary.
       */
      if (p_ref->p_event_queue == NULL)
         return;

      /* Disconnect the linkage of the age reference within the event queue
       * leaving the event queue intact with this element removed.
       */
      if (p_ref->p_prev == NULL)
         p_ref->p_event_queue->p_head = p_ref->p_next;
      else
         p_ref->p_prev->p_next = p_ref->p_next;

      if (p_ref->p_next == NULL)
         {
         /* If the element being removed is at the tail of the event queue,
          * the total time of the event queeu must be decreased by that
          * element's delta time.  Otherwise the total time of the event
          * queue must not be changed.
          */
         p_ref->p_event_queue->total_time -= p_ref->delta_time;
         p_ref->p_event_queue->p_tail      = p_ref->p_prev;
         }
      else
         {
         /* Because the event queue contains a delta time for each element,
          * the removal of an element requires that the delta time of the
          * next entry be adjusted to maintain the desired absolute times
          * for each event.  This is done by adding the delta time of the
          * removed entry to the entry that previously followed the removed
          * entry.
          */
         p_ref->p_next->delta_time += p_ref->delta_time;
         p_ref->p_next->p_prev      = p_ref->p_prev;
         }

      /* Adjust the size of the event queue. */

      p_ref->p_event_queue->size -= 1;

      /* Set the event queue pointer in the reference to NULL. */

      p_ref->p_event_queue = NULL;
      }
   }

/****************************************************************************
 *  Clock Callback Handlers
 ****************************************************************************/

static BOOLEAN esi_clock_update (HANDLE esi_handle,
                                 HANDLE la_elan_handle)
   {
   LA_CONTEXT        *p_context;
   MAP_ESI_RECORD    *p_esi;
   LA_ELAN_CONTEXT   *p_elan;
   UINT32             time;
#ifdef PROXY
   UINT8              vpi;
   UINT16             vci;
#endif

   p_esi     = (MAP_ESI_RECORD  *) esi_handle;
   p_elan    = (LA_ELAN_CONTEXT *) la_elan_handle;
   p_context = (LA_CONTEXT      *) p_elan->la_handle;

   /* If The ESI record is not linked into an event queue, set its status
    * to "UNKNOWN".  ESI records are normally always linked into an event
    * queue so they can be refreshed before they go stale.  If an ESI record
    * is still of legal age but has given up retrying ARPs (which can happen
    * if a peer LEC disconnects from the network), the entry will be left
    * unattached to an event queue.  In this case, we wish for the LE-ARP
    * protocol to be restarted the next time a packet is sent to this ESI.
    */
   if (p_esi->p_age_ref->p_event_queue == NULL)
      p_esi->esi_status = ESI_UNKNOWN;

   /* Process age values. */

   p_esi->arp_age += 1000;
   p_esi->use_age += 1000;

   /* If the age has reached the limit, delete the current entry. */

   if ((p_esi->remote_flag) && (p_elan->topology_change_flag))
      time = p_elan->forward_delay_time;
   else
      time = p_elan->aging_time;

   if (((p_esi->arp_age >= time) || (p_esi->use_age >= USE_AGE_LIMIT)) &&
       (!p_esi->permanent_entry))
      {
      esi_linkage_remove ((HANDLE) p_esi);
      map_element_delete (p_elan->esi_list, (HANDLE) p_esi);
      os_mem_dealloc (p_context->os_handle, p_esi->p_age_ref);
      os_mem_dealloc (p_context->os_handle, p_esi);
      }
   else
      {
      /* Process tick_count. */

      p_esi->tick_count += 1;
      if (((p_esi->esi_status == ESI_VC_PENDING) ||
           (p_esi->esi_status == ESI_ARP_PENDING)) &&
          (p_esi->tick_count >= p_elan->max_unknown_frame_time))
         {
         p_esi->tick_count      = 0;
         p_esi->packets_flooded = 0;
         }

      if ((p_esi->esi_status == ESI_FLUSH_PENDING) &&
          (p_esi->tick_count >= p_elan->path_switching_delay))
         {
         p_esi->esi_status = ESI_FORWARD_DIRECT;
#ifdef PROXY
         conn_vcid_get ((HANDLE) p_esi->vc_link, &vpi, &vci);
         proxy_data_direct_add (p_elan->lport_handle,
                                p_esi->esi,
                                vpi,
                                vci,
                                PROXY_FORMAT_802_3,  /* Temporary */
                                (HANDLE) p_esi->vc_link);
#endif
         }
      }

   return TRUE;
   }


/*++
* =====================
* = la_clock_callback =
* =====================
*
* Overview:
*   This is the one-second clock handler.  It is called by the operating
*   system once every second.
*
* Arguments:
*   la_handle  - (IN)  Handle for this LEC_ARP instance.  Provides
*                      the proper context in which to run.
*              
* Returns:
*   None
*
* Preconditions:
*   None
*
* Postconditions:
*   None
*
* Description:
*   This handler performs the periodic tasks that are required by the mapping.
*   It is not concerned with LE-ARP retries or refreshes.  It must perform
*   the following:
*
*       1.  Re-set the timer for the next one-second period.
*
*       2.  Update the time_since_creation counter.
*
*       3.  For each ESI in the LEARP-Cache, increment the arp_age and
*           delete entries that have aged past the age limit.
*
*       4.  For each ESI in the LEARP-Cache, increment the tick_count
*           and depending on the esi_status of the entry, perform the required
*           maintenance:
*
*           a.  If the esi_status is ESI_VC_PENDING or ESI_ARP_PENDING then
*               if the tick count is above the max_unknown_frame_time, clear
*               the tick_count to zero and clear the packets_flooded counter
*               to zero.  This supports the packet rate limit per address
*               while flooding unknowns.
*
*           b.  If the esi_status is ESI_FLUSH_PENDING and the tick_count is
*               greater than or equal to the path_switching_delay, change the
*               esi_status to ESI_FORWARD_DIRECT.  This causes the flush period
*               to end regardless of the progress of the flush protocol.
*
* Issues:
*
--*/
STATIC void la_clock_callback (HANDLE la_handle)
   {
   LA_CONTEXT        *p_context;
   LA_ELAN_CONTEXT   *p_elan;

   p_context = (LA_CONTEXT*) la_handle;

   /* Restart timer for the next one-second period. */

   os_timer_set (p_context->one_sec_clock, 1000 /* msec */);

   /* Increment overall life timer. */

   p_context->time_since_creation += 1;

   /* Process the contents of each LEC. */

   utl_list_traverse (p_context->elan_list, p_elan)
      {
      map_list_traverse (p_elan->esi_list, esi_clock_update, (HANDLE) p_elan);
      }
   }

static void process_event_queue (EVENT_QUEUE       *p_event_queue,
                                 LA_CONTEXT        *p_context,
                                 LA_ELAN_CONTEXT   *p_elan)
   {
   MAP_ESI_RECORD    *p_esi;
   AGE_REF           *p_age_ref;
   UINT32             interval_time;

   /* First process the fixed event queue. */

   interval_time = REFRESH_INTERVAL;

   /* The following loop is necessary because there may be a string of
    * age references that are due to be serviced during this interval.
    */
   event_queue_remove (p_event_queue,
                       &interval_time,
                       &p_age_ref);
   while (p_age_ref != NULL)
      {
      p_esi = (MAP_ESI_RECORD *) p_age_ref->esi_handle;

      /* Service the event. */

      if (p_context->arp_xmt_callback != NULL)
         p_context->arp_xmt_callback (p_elan->lc_elan_handle, p_esi->esi);

      /* Queue a retry event. */

      p_age_ref->arp_retries += 1;

      if (p_age_ref->arp_retries <= p_elan->max_retry_count)
         {
         event_queue_insert (&p_elan->fixed_event_queue,
                             p_age_ref,
                             p_elan->le_arp_response_time);
         }

      event_queue_remove (p_event_queue,
                          &interval_time,
                          &p_age_ref);
      }
   }

/*++
* ===========================
* = la_arp_refresh_callback =
* ===========================
*
* Overview:
*   This is the arp refresh callback handler.  This is a separate handler
*   from the one-second clock because the arp refresh will most likely be
*   handled at a slightly higher frequency than one-per-second.
*
* Arguments:
*   la_handle  - (IN)  Handle for this LEC_ARP instance.  Provides
*                      the proper context in which to run.
*              
* Returns:
*   None
*
* Preconditions:
*   None
*
* Postconditions:
*   None
*
* Description:
*   This handler services the two event queues in each LA_ELAN record.  When an
*   age reference is serviced from either event queue, an LE-ARP request is
*   triggered.
*
* Issues:
*
--*/
STATIC void la_arp_refresh_callback (HANDLE la_handle)
   {
   LA_CONTEXT        *p_context;
   LA_ELAN_CONTEXT   *p_elan;

   p_context = (LA_CONTEXT *) la_handle;

   /* Restart timer for the next period. */

   (void) os_timer_set (p_context->arp_refresh_clock, REFRESH_INTERVAL);

   /* Process each LEC. */

   utl_list_traverse (p_context->elan_list, p_elan)
      {
      process_event_queue (&p_elan->fixed_event_queue,    p_context, p_elan);
      process_event_queue (&p_elan->variable_event_queue, p_context, p_elan);
      }
   }


/****************************************************************************
 *  Public Functions - Refer to lec_arp.h for interface specifications.
 ****************************************************************************/

/*++
* =============
* = la_create =
* =============
*
* Status:      IMPLEMENTATION COMPLETE
*
* Description:
*
* Issues:
*
--*/
STATUS la_create (HANDLE                os_handle,
                  HANDLE                lc_handle,
                  HANDLE                cm_handle,
                  ARP_XMT_CALLBACK      arp_xmt_callback,
                  FLUSH_XMT_CALLBACK    flush_xmt_callback,
                  SVC_SETUP_CALLBACK    svc_setup_callback,
                  HANDLE               *p_la_handle)
   {
   LA_CONTEXT     *p_context;
   STATUS          status;

   p_context = (LA_CONTEXT *) os_mem_alloc (os_handle, sizeof (LA_CONTEXT));
   if (p_context == NULL)
      return STATUS_K_RESOURCES;

   status = os_timer_alloc (os_handle,
                            la_clock_callback,
                            (HANDLE) p_context,
                            &(p_context->one_sec_clock));
   if (status != STATUS_K_SUCCESS)
      {
      os_mem_dealloc (os_handle, p_context);
      return STATUS_K_RESOURCES;
      }

   status = os_timer_alloc (os_handle,
                            la_arp_refresh_callback,
                            (HANDLE) p_context,
                            &(p_context->arp_refresh_clock));
   if (status != STATUS_K_SUCCESS)
      {
      os_timer_dealloc (p_context->one_sec_clock);
      os_mem_dealloc (os_handle, p_context);
      return STATUS_K_RESOURCES;
      }

   cm_map_get (cm_handle,
               link_update_callback,
               link_remove_callback,
               (HANDLE) p_context,
               &p_context->vc_list,
               &p_context->atm_list);

   p_context->os_handle           = os_handle;
   p_context->lc_handle           = lc_handle;
   p_context->cm_handle           = cm_handle;
   p_context->arp_xmt_callback    = arp_xmt_callback;
   p_context->flush_xmt_callback  = flush_xmt_callback;
   p_context->svc_setup_callback  = svc_setup_callback;
   p_context->time_since_creation = 0;

   utl_list_init (p_context->elan_list);

   os_timer_set (p_context->one_sec_clock,     1000);
   os_timer_set (p_context->arp_refresh_clock, REFRESH_INTERVAL);

   *p_la_handle = (HANDLE) p_context;

   return STATUS_K_SUCCESS;
   }

/*++
* ==============
* = la_destroy =
* ==============
*
* Status:      IMPLEMENTATION COMPLETE
*
* Description:
*
* Issues:
*
--*/
void la_destroy (HANDLE  la_handle)
   {
   LA_CONTEXT        *p_context;
   LA_ELAN_CONTEXT   *p_elan;

   p_context = (LA_CONTEXT *) la_handle;

   /* Deregister all ELAN contexts */

   utl_list_traverse (p_context->elan_list, p_elan)
      {
      la_deregister ((HANDLE) p_elan);
      }

   /* Deallocate the timers. */

   os_timer_dealloc (p_context->one_sec_clock);
   os_timer_dealloc (p_context->arp_refresh_clock);

   /* Deallocate the memory used for the LA context. */

   os_mem_dealloc (p_context->os_handle, p_context);
   }

/*++
* ===============
* = la_register =
* ===============
*
* Status:      IMPLEMENTATION COMPLETE
*
* Description:
*
* Issues:
*
--*/
STATUS la_register (HANDLE    la_handle,
                    HANDLE    lc_elan_handle,
                    HANDLE    lport_handle,
                    char     *p_text,
                    HANDLE   *p_la_elan_handle)
   {
   LA_CONTEXT        *p_context;
   LA_ELAN_CONTEXT   *p_elan;
   STATUS             status;

   /* Obtain a pointer to the LA Context block. */

   p_context = (LA_CONTEXT *) la_handle;

   /* Allocate space for the LA_ELAN context block.  If there is a problem,
    * return error status.
    */
   p_elan = (LA_ELAN_CONTEXT *) os_mem_alloc (p_context->os_handle,
                                              sizeof (LA_ELAN_CONTEXT));
   if (p_elan == NULL)
      return STATUS_K_RESOURCES;

   /* Initialize the fields of the LA_ELAN context block.  Mark the ELAN as
    * unconfigured.
    */
   p_elan->la_handle      = la_handle;
   p_elan->lc_elan_handle = lc_elan_handle;
   p_elan->lport_handle   = lport_handle;
   p_elan->configured     = FALSE;
   p_elan->p_text         = p_text;
   p_elan->event_mask     = EM_DEFAULT;
   p_elan->cache_size     = 0;

   event_queue_init (&p_elan->fixed_event_queue);
   event_queue_init (&p_elan->variable_event_queue);

   /* Initialize the ESI list in the mapping.  If there is a problem with the
    * allocation, return error status.
    */
   status = map_list_alloc (p_context->os_handle,
                            sizeof (ESI),
                            &p_elan->esi_list);
   if (status != STATUS_K_SUCCESS)
      {
      os_mem_dealloc (p_context->os_handle, p_elan);
      return STATUS_K_RESOURCES;
      }

   utl_list_add (p_context->elan_list, p_elan);

   *p_la_elan_handle = (HANDLE) p_elan;
   return STATUS_K_SUCCESS;
   }

static BOOLEAN la_deregister_handler (HANDLE esi_handle,
                                      HANDLE la_handle)
   {
   LA_CONTEXT     *p_context;
   MAP_ESI_RECORD *p_esi;

   p_context = (LA_CONTEXT     *) la_handle;
   p_esi     = (MAP_ESI_RECORD *) esi_handle;

   esi_linkage_remove (esi_handle);
   age_ref_unlink (p_esi->p_age_ref);

   os_mem_dealloc (p_context->os_handle, p_esi->p_age_ref);
   os_mem_dealloc (p_context->os_handle, p_esi);

   return TRUE;
   }

/*++
* =================
* = la_deregister =
* =================
*
* Status:
*
* Description:
*
* Issues:
*
--*/
void la_deregister (HANDLE la_elan_handle)
   {
   LA_CONTEXT        *p_context;
   LA_ELAN_CONTEXT   *p_elan;

   p_elan    = (LA_ELAN_CONTEXT *) la_elan_handle;
   p_context = (LA_CONTEXT      *) p_elan->la_handle;

   map_list_dealloc (p_elan->esi_list,
                     la_deregister_handler,
                     p_elan->la_handle);

   assert (p_elan->fixed_event_queue.size    == 0);
   assert (p_elan->variable_event_queue.size == 0);

   utl_list_delete (p_context->elan_list, p_elan);

   os_mem_dealloc (p_context->os_handle, p_elan);
   }

/*++
* =====================
* = la_event_mask_set =
* =====================
*
* Status:      IMPLEMENTATION COMPLETE
*
* Description:
*
* Issues:
*
--*/
void la_event_mask_set (HANDLE la_elan_handle,
                        UINT32 event_mask)
   {
   LA_ELAN_CONTEXT   *p_elan;

   p_elan = (LA_ELAN_CONTEXT *) la_elan_handle;

   p_elan->event_mask = event_mask;
   }

/*++
* =============
* = la_config =
* =============
*
* Status:      IMPLEMENTATION COMPLETE
*
* Description:
*
* Issues:
*
--*/
void la_config (HANDLE     la_elan_handle,
                LAN_TYPE   lan_type,
                BOOLEAN    proxy_flag,
                UINT16     max_unknown_frame_count,
                UINT16     max_unknown_frame_time,
                UINT32     vcc_timeout,
                UINT16     max_retry_count,
                UINT32     aging_time,
                UINT16     forward_delay_time,
                UINT16     le_arp_response_time,
                UINT16     flush_timeout,
                UINT16     path_switching_delay)
   {
   LA_ELAN_CONTEXT   *p_elan;

   p_elan = (LA_ELAN_CONTEXT *) la_elan_handle;

   p_elan->configured               = TRUE;
   p_elan->lan_type                 = lan_type;
   p_elan->proxy_flag               = proxy_flag;
   p_elan->topology_change_flag     = FALSE;
   p_elan->max_unknown_frame_count  = max_unknown_frame_count;
   p_elan->max_unknown_frame_time   = max_unknown_frame_time;
   p_elan->vcc_timeout              = vcc_timeout;
   p_elan->max_retry_count          = max_retry_count;
   p_elan->aging_time               = aging_time * 1000;
   p_elan->forward_delay_time       = forward_delay_time * 1000;
   p_elan->le_arp_response_time     = le_arp_response_time * 1000;
   p_elan->flush_timeout            = flush_timeout;
   p_elan->path_switching_delay     = path_switching_delay;
   }


static BOOLEAN reset_handler (HANDLE   esi_handle,
                              HANDLE   la_elan_handle)
   {
   LA_CONTEXT        *p_context;
   LA_ELAN_CONTEXT   *p_elan;
   MAP_ESI_RECORD    *p_esi;

   p_elan    = (LA_ELAN_CONTEXT *) la_elan_handle;
   p_context = (LA_CONTEXT      *) p_elan->la_handle;

   /* Obtain a pointer to the ESI record structure. */

   p_esi = (MAP_ESI_RECORD *) esi_handle;

   /* Remove only non-permanent entries from the cache. */

   if (!p_esi->permanent_entry)
      {
      /* Remove any back links from ATM address records to this ESI record. */

      esi_linkage_remove (esi_handle);

      /* Remove the ESI record from the esi_list. */

      map_element_delete (p_elan->esi_list, p_esi);

      /* Free the memory used to store the ESI record. */

      os_mem_dealloc (p_context->os_handle, p_esi->p_age_ref);
      os_mem_dealloc (p_context->os_handle, p_esi);
      }

   return TRUE;
   }

/*++
* =================
* = la_elan_reset =
* =================
*
* Status:
*
* Description:
*
* Issues:
*
--*/
void la_elan_reset (HANDLE la_elan_handle)
   {
   LA_ELAN_CONTEXT   *p_elan;

   p_elan    = (LA_ELAN_CONTEXT *) la_elan_handle;

   map_list_traverse (p_elan->esi_list, reset_handler, la_elan_handle);
   }

/*++
* ==================
* = la_arp_resolve =
* ==================
*
* Status:
*
* Description:
*
* Issues:
*
--*/
void la_arp_resolve (HANDLE        la_elan_handle,
                     ESI           esi,
                     XMT_ACTION   *p_action,
                     HANDLE       *p_conn_handle)
   {
   LA_ELAN_CONTEXT  *p_elan;
   LA_CONTEXT       *p_context;
   MAP_ESI_RECORD   *p_esi;
   STATUS            status;
   HANDLE            esi_handle;
   HANDLE            unused_atm_handle;
   ADDR_ATM          unused_addr_atm;

   /* Obtain a pointer to the ELAN context block. */

   p_elan = (LA_ELAN_CONTEXT *) la_elan_handle;

   /* Search the ARP Cache for a record matching the esi. */

   status = map_key_search (p_elan->esi_list, esi, &esi_handle);

   if (status == STATUS_K_SUCCESS)
      {
      /* If a matching record was found... */

      /* Convert the esi_handle to a pointer to the esi record. */

      p_esi = (MAP_ESI_RECORD *) esi_handle;

      /* Set the use_age to zero since this entry is now being used. */

      p_esi->use_age = 0;

      /* If the esi_status in the record is ESI_FORWARD_DIRECT, the data
       * direct VC is set up and ready to use, return the connection handle of
       * the data direct VC and an action of XMT_FORWARD.
       */
      if (p_esi->esi_status == ESI_FORWARD_DIRECT)
         {
         *p_conn_handle = p_esi->vc_link;
         *p_action      = XMT_FORWARD;
         return;
         }

      /* If the esi_status is "UNKNOWN", begin the LE-ARP process to
       * reestablish a data-direct VCC for this ESI.  Then fall through
       * to the flooding code below.
       */
      if (p_esi->esi_status == ESI_UNKNOWN)
         {
         /* Call the CTRL module and request an ARP packet for this address. */

         p_context = (LA_CONTEXT *) p_elan->la_handle;
         p_esi->esi_status = ESI_ARP_PENDING;
         p_context->arp_xmt_callback (p_elan->lc_elan_handle, esi);

         /* Place the reference to this esi record on the fixed event queue
          * for a retry.
          */
         p_esi->p_age_ref->arp_retries = 1;
         event_queue_insert (&p_elan->fixed_event_queue,
                             p_esi->p_age_ref,
                             p_elan->le_arp_response_time);
         }

      /* If the data direct VC is not yet set up, check to see if the unknown
       * frame count is greater than the limit.  If the limit has not been
       * reached, allow the caller to flood the packet.
       */
      if (p_esi->packets_flooded < p_elan->max_unknown_frame_count)
         {
         p_esi->packets_flooded += 1;
         *p_action               = XMT_FLOOD;
         return;
         }

      /* If flooding is not allowed, tell the caller to discard the packet. */

      *p_action = XMT_DISCARD;
      return;
      }
   else
      {
      /* If no matching record was found...
       *
       * Create a new esi record to be placed in the arp cache.  If there is
       * an error allocating memory, simply tell the caller to discard the
       * packet and return.  There is nothing else to do here.
       */
      p_context = (LA_CONTEXT *)     p_elan->la_handle;

      status = esi_record_make (la_elan_handle,
                                esi,
                                FALSE,
                                unused_addr_atm,
                                &esi_handle,
                                &unused_atm_handle);
      if (status != STATUS_K_SUCCESS)
         {
         *p_action = XMT_DISCARD;
         return;
         }

      p_esi = (MAP_ESI_RECORD *) esi_handle;

      p_esi->esi_status          = ESI_ARP_PENDING;
      p_esi->packets_flooded     = 1;
      p_esi->tick_count          = 0;
      p_esi->remote_flag         = TRUE;    /* Make this assumption */
      p_esi->permanent_entry     = FALSE;

      /* Call the CTRL module and request an ARP packet for this address. */

      p_context->arp_xmt_callback (p_elan->lc_elan_handle, esi);

      /* Place the reference to this esi record on the fixed event queue
       * for a retry.
       */
      p_esi->p_age_ref->arp_retries = 1;
      event_queue_insert (&p_elan->fixed_event_queue,
                          p_esi->p_age_ref,
                          p_elan->le_arp_response_time);

      /* Allow the caller to flood the packet. */

      if (p_elan->max_unknown_frame_count == 0)
         *p_action = XMT_DISCARD;
      else
         *p_action = XMT_FLOOD;
      return;
      }
   }

/*++
* ===============
* = la_perm_add =
* ===============
*
* Status:
*
* Description:
*
* Issues:
*
--*/
STATUS la_perm_add (HANDLE    la_elan_handle,
                    ESI       esi,
                    ADDR_ATM  addr_atm)
   {
   LA_CONTEXT        *p_context;
   LA_ELAN_CONTEXT   *p_elan;
   MAP_ESI_RECORD    *p_esi;
   MAP_ATM_RECORD    *p_atm;
   HANDLE             esi_handle;
   HANDLE             atm_handle;
   STATUS             status;
#ifdef PROXY
   UINT8              vpi;
   UINT16             vci;
#endif

   p_elan    = (LA_ELAN_CONTEXT *) la_elan_handle;
   p_context = (LA_CONTEXT *)      p_elan->la_handle;

   status = esi_record_make (la_elan_handle,
                             esi,
                             TRUE,
                             addr_atm,
                             &esi_handle,
                             &atm_handle);
   if (status != STATUS_K_SUCCESS)
      return STATUS_K_RESOURCES;

   p_esi = (MAP_ESI_RECORD *) esi_handle;
   p_atm = (MAP_ATM_RECORD *) atm_handle;

   /* Initialize the fields of the ESI record. */

   p_esi->permanent_entry = TRUE;

   switch (p_atm->atm_status)
      {
      case ATM_NONE:

         /* An SVC must be requested for this address. */

         p_atm->atm_status = ATM_PENDING;
         p_esi->esi_status = ESI_VC_PENDING;
         p_context->svc_setup_callback (p_elan->lc_elan_handle, addr_atm);
         break;

      case ATM_PENDING:

         p_esi->esi_status = ESI_VC_PENDING;
         break;

      case ATM_VALID:

         p_esi->esi_status = ESI_FORWARD_DIRECT;
#ifdef PROXY
         conn_vcid_get ((HANDLE) p_esi->vc_link, &vpi, &vci);
         proxy_data_direct_add (p_elan->lport_handle,
                                p_esi->esi,
                                vpi,
                                vci,
                                PROXY_FORMAT_802_3,  /* Temporary */
                                (HANDLE) p_esi->vc_link);
#endif
         break;

      default:
         assert (FALSE);
      }

   return STATUS_K_SUCCESS;
   }

/*++
* ==================
* = la_addr_delete =
* ==================
*
* Status:   IMPLEMENTATION COMPLETE
*
* Description:
*
* Issues:
*
--*/
STATUS la_addr_delete (HANDLE    la_elan_handle,
                       ESI       esi,
                       BOOLEAN   permanent)
   {
   LA_CONTEXT        *p_context;
   LA_ELAN_CONTEXT   *p_elan;
   MAP_ESI_RECORD    *p_esi;
   STATUS             status;
   HANDLE             esi_handle;

   p_elan    = (LA_ELAN_CONTEXT *) la_elan_handle;
   p_context = (LA_CONTEXT      *) p_elan->la_handle;

   /* Search the esi_list for a matching entry.  Return NOT_FOUND if there
    * is no match.
    */
   status = map_key_search (p_elan->esi_list, esi, &esi_handle);
   if (status != STATUS_K_SUCCESS)
      return STATUS_K_NOT_FOUND;

   /* Obtain a pointer to the ESI record structure. */

   p_esi = (MAP_ESI_RECORD *) esi_handle;

   /* If the permanent flag is not set and this is a permanent entry, return
    * NOT_FOUND.  This case occurs when a NACK is received for an ESI for which
    * there is a permanent LE_ARP cache entry.
    */
   if (!permanent && (p_esi->permanent_entry))
      return STATUS_K_NOT_FOUND;

   /* Remove any back links from ATM address records to this ESI record. */

   esi_linkage_remove (esi_handle);

   /* Remove the ESI record from the esi_list. */

   map_element_delete (p_elan->esi_list, p_esi);

   /* Free the memory used to store the ESI record. */

   os_mem_dealloc (p_context->os_handle, p_esi);

   return STATUS_K_SUCCESS;
   }

/*++
* =================
* = la_arp_update =
* =================
*
* Status:
*
* Description:
*
* Issues:
*
--*/
void la_arp_update (HANDLE    la_elan_handle,
                    ESI       esi,
                    ADDR_ATM  addr_atm,
                    BOOLEAN   remote_flag)
   {
   LA_CONTEXT        *p_context;
   LA_ELAN_CONTEXT   *p_elan;
   MAP_ESI_RECORD    *p_esi;
   MAP_ATM_RECORD    *p_atm;
   HANDLE             esi_handle;
   HANDLE             atm_handle;
   STATUS             status;
   UINT32             time;
#ifdef PROXY
   UINT8              vpi;
   UINT16             vci;
#endif

   p_elan    = (LA_ELAN_CONTEXT *) la_elan_handle;
   p_context = (LA_CONTEXT      *) p_elan->la_handle;

   status = map_key_search (p_elan->esi_list, esi, &esi_handle);

   if (status == STATUS_K_SUCCESS)
      {
      p_esi = (MAP_ESI_RECORD *) esi_handle;

      p_esi->arp_age = 0;

      if (p_esi->permanent_entry)
         return;

      p_esi->remote_flag = remote_flag;

      status = esi_atm_link (la_elan_handle,
                             (HANDLE) p_esi,
                             addr_atm,
                             &atm_handle);

      if (status == STATUS_K_SUCCESS)
         {
         p_atm = (MAP_ATM_RECORD *) atm_handle;

         switch (p_atm->atm_status)
            {
            case ATM_NONE:
               /* An SVC must be requested for this address. */

               p_context->svc_setup_callback (p_elan->lc_elan_handle,
                                              addr_atm);
               p_atm->atm_status = ATM_PENDING;
               p_esi->esi_status = ESI_VC_PENDING;
               break;
            case ATM_PENDING:
               p_esi->esi_status = ESI_VC_PENDING;
               break;
            case ATM_VALID:
               /* 
                */
               if ((p_esi->esi_status == ESI_VC_PENDING) ||
                   (p_esi->esi_status == ESI_ARP_PENDING))
                  {
                  if (p_elan->max_unknown_frame_count == 0)
           {
                     p_esi->esi_status = ESI_FORWARD_DIRECT;
#ifdef PROXY
                     conn_vcid_get ((HANDLE) p_esi->vc_link, &vpi, &vci);
                     proxy_data_direct_add (p_elan->lport_handle,
                                            p_esi->esi,
                                            vpi,
                                            vci,
                                            PROXY_FORMAT_802_3,  /* Temporary */
                                            (HANDLE) p_esi->vc_link);
#endif
           }
                  else
                     {
                     p_esi->esi_status = ESI_FLUSH_PENDING;
                     p_context->flush_xmt_callback (p_elan->lc_elan_handle,
                                                    p_atm->addr,
                                                    p_esi);
                     }
                  }
               break;
            default:
               assert (FALSE);
            }

         /* Remove the ESI reference from the event queue and re-queue it
          * on the appropriate event queue.
          */
         age_ref_unlink (p_esi->p_age_ref);
         p_esi->p_age_ref->arp_retries = 0;
         if (p_esi->remote_flag)
            {
            time = p_elan->topology_change_flag ?
                        p_elan->forward_delay_time :
                        p_elan->aging_time;

            event_queue_insert (&p_elan->variable_event_queue,
                                p_esi->p_age_ref,
                                time - (p_elan->max_retry_count *
                                        p_elan->le_arp_response_time));
            }
         else
            {
            event_queue_insert (&p_elan->fixed_event_queue,
                                p_esi->p_age_ref,
                                p_elan->aging_time -
               (p_elan->max_retry_count *
                p_elan->le_arp_response_time));
            }
         }
      }
   }

/*++
* =====================
* = la_flush_complete =
* =====================
*
* Status:
*
* Description:
*
* Issues:
*
--*/
void la_flush_complete (HANDLE   la_elan_handle,
                        ADDR_ATM addr_atm,
                        UINT32   tran_id)
   {
   LA_CONTEXT        *p_context;
   LA_ELAN_CONTEXT   *p_elan;
   HANDLE             atm_handle;
   MAP_ATM_RECORD    *p_atm;
   MAP_ESI_RECORD    *p_esi;
   STATUS             status;
   HANDLE_RECORD     *p_handle;
#ifdef PROXY
   UINT8              vpi;
   UINT16             vci;
#endif

   /* Obtain the structure pointers that will be needed in this routine. */

   p_elan    = (LA_ELAN_CONTEXT *) la_elan_handle;
   p_context = (LA_CONTEXT      *) p_elan->la_handle;

   /* Perform a lookup on the atm_address in the atm_list. */

   status = map_key_search (p_context->atm_list, &addr_atm, &atm_handle);

   /* If the search was successful, continue processing the flush complete
    * indication.  If the search was not successful, conclude that the flush
    * response was spurious and ignore it.
    */
   if (status == STATUS_K_SUCCESS)
      {
      /* Obtain a pointer to the atm address record. */

      p_atm = (MAP_ATM_RECORD *) atm_handle;

      /* Process each of the atm address record's associated ESI back links. */

      utl_list_traverse (p_atm->esi_link_list, p_handle)
         {
         p_esi = (MAP_ESI_RECORD *) p_handle->handle;
         if (p_esi->flush_tran_id  == tran_id)
            {
            assert (p_atm->atm_status == ATM_VALID);
            p_esi->esi_status = ESI_FORWARD_DIRECT;
#ifdef PROXY
            conn_vcid_get ((HANDLE) p_esi->vc_link, &vpi, &vci);
            proxy_data_direct_add (p_elan->lport_handle,
                                   p_esi->esi,
                                   vpi,
                                   vci,
                                   PROXY_FORMAT_802_3,  /* Temporary */
                                   (HANDLE) p_esi->vc_link);
#endif
            }
         }
      }
   }

/*++
* ==========================
* = la_topology_change_set =
* ==========================
*
* Status:
*
* Description:
*
* Issues:
*
--*/
void la_topology_change_set (HANDLE  la_elan_handle,
                             BOOLEAN topology_change_flag)
   {
   LA_ELAN_CONTEXT   *p_elan;
   MAP_ESI_RECORD    *p_esi;
   AGE_REF           *p_age_ref;
   UINT32             total, factor;
   BOOLEAN            multiply_flag;

   p_elan = (LA_ELAN_CONTEXT *) la_elan_handle;

   if (topology_change_flag == p_elan->topology_change_flag)
      return;

   p_elan->topology_change_flag = topology_change_flag;

   /* Calculate the relationship between the old and the new.  Determine if
    * multiplication or division is required.
    */
   factor = p_elan->aging_time / p_elan->forward_delay_time;
   multiply_flag = !topology_change_flag;

   /* Adjust all of the delta times in the variable event queue. */

   p_age_ref = p_elan->variable_event_queue.p_head;
   total     = 0;
   while (p_age_ref != NULL)
      {
      p_esi = (MAP_ESI_RECORD *) p_age_ref->esi_handle;
      if (multiply_flag)
         {
         p_age_ref->delta_time *= factor;
         p_esi->arp_age        *= factor;
         }
      else
         {
         p_age_ref->delta_time /= factor;
         p_esi->arp_age        /= factor;
         }

      total    += p_age_ref->delta_time;
      p_age_ref = p_age_ref->p_next;
      }

   p_elan->variable_event_queue.total_time = total;
   }

/*++
* ===============
* = la_vc_added =
* ===============
*
* Status:
*
* Description:
*
* Issues:
*
--*/
void la_vc_added (HANDLE la_elan_handle,
                  HANDLE conn_handle)
   {
   LA_CONTEXT        *p_context;
   LA_ELAN_CONTEXT   *p_elan;
   MAP_VC_RECORD     *p_vc;
   MAP_ATM_RECORD    *p_atm;
   MAP_ESI_RECORD    *p_esi;
   HANDLE_RECORD     *p_handle;
#ifdef PROXY
   UINT8              vpi;
   UINT16             vci;
#endif

   p_elan    = (LA_ELAN_CONTEXT *) la_elan_handle;
   p_context = (LA_CONTEXT      *) p_elan->la_handle;
   p_vc      = (MAP_VC_RECORD   *) conn_handle;
   p_atm     = (MAP_ATM_RECORD  *) p_vc->atm_link;

   utl_list_traverse (p_atm->esi_link_list, p_handle)
      {
      p_esi = (MAP_ESI_RECORD *) p_handle->handle;
      if (p_esi->esi_status == ESI_VC_PENDING)
         {
         if (p_elan->max_unknown_frame_count == 0)
       {
            p_esi->esi_status = ESI_FORWARD_DIRECT;
#ifdef PROXY
            conn_vcid_get (conn_handle, &vpi, &vci);
            proxy_data_direct_add (p_elan->lport_handle,
                                   p_esi->esi,
                                   vpi,
                                   vci,
                                   PROXY_FORMAT_802_3,  /* Temporary */
                                   conn_handle);
#endif
       }
         else
            {
            p_esi->esi_status = ESI_FLUSH_PENDING;
            p_context->flush_xmt_callback (p_elan->lc_elan_handle,
                                           p_atm->addr,
                                           (HANDLE) p_esi);
            }
         }
      }
   }


/*++
* ===============
* = la_attr_get =
* ===============
*
* Status:
*
* Description:
*
* Issues:
*
--*/
STATUS la_attr_get (HANDLE     la_elan_handle,
                    UINT16     attr_id,
                    void      *p_out_buffer,
                    UINT16     out_buffer_length,
                    UINT16    *p_bytes_written)
   {
   LA_CONTEXT        *p_context;
   LA_ELAN_CONTEXT   *p_elan;
   UINT16             length;
   UINT8             *p_value;
   int                i;

   p_elan    = (LA_ELAN_CONTEXT *) la_elan_handle;
   p_context = (LA_CONTEXT      *) p_elan->la_handle;

   switch (attr_id)
      {
      case ATTR_C16_LEARP_CACHE:
         break;

      default:
         return STATUS_K_ARG_2_BAD;
      }

   if (out_buffer_length < length)
      return STATUS_K_ARG_5_BAD;

   *p_bytes_written = length;
   for (i = 0; i < length; i++)
      ((UINT8 *) p_out_buffer)[i] = p_value[i];
   return STATUS_K_SUCCESS;
   }

STATUS la_attr_get_next (HANDLE     la_elan_handle,
                         UINT16     attr_id,
                         void      *p_in_buffer,
                         UINT16     in_buffer_length,
                         void      *p_out_buffer,
                         UINT16     out_buffer_length,
                         UINT16    *p_bytes_written)
   {
   LA_CONTEXT        *p_context;
   LA_ELAN_CONTEXT   *p_elan;
   UINT16             length;
   UINT8             *p_value;
   int                i;

   p_elan    = (LA_ELAN_CONTEXT *) la_elan_handle;
   p_context = (LA_CONTEXT      *) p_elan->la_handle;

   switch (attr_id)
      {
      case ATTR_C16_LEARP_CACHE:
         break;

      default:
         return STATUS_K_ARG_2_BAD;
      }

   if (out_buffer_length < length)
      return STATUS_K_ARG_5_BAD;

   *p_bytes_written = length;
   for (i = 0; i < length; i++)
      ((UINT8 *) p_out_buffer)[i] = p_value[i];
   return STATUS_K_SUCCESS;
   }

/*++
* ===============
* = la_attr_set =
* ===============
*
* Status:
*
* Description:
*
* Issues:
*
--*/
STATUS la_attr_set (HANDLE     la_elan_handle,
                    UINT16     attr_id,
                    void      *p_in_buffer,
                    UINT16     in_buffer_length)
   {
   return STATUS_K_NO_PRIVILEGE;
   }

