/*************************************************************************
  (c) Copyright.  Digital Equipment Corporation, 1995.  All Rights
  Reserved.

  Permission is hereby granted to use, copy, modify, or enhance this 
  software freely, as long as the foregoing copyright of Digital Equipment
  Corporation and this notice are retained on the software.  This 
  software may not be distributed or sublicensed for a fee.  Digital      
  makes this software available "AS IS" and without warranties of any
  kind.  
 *************************************************************************/

/*
* Module Name:
*   utl.c
*   
* Abstract:
*   This file implements the general utilities.
*                                                                             
* Authors:
*   DMW - Douglas M. Washabaugh
*   TLR - Theodore L. Ross
*
* Modification History:
*   Date       Name  Description 
*   08-Aug-94  DMW   - Created.
*   30-Sep-94  ND    - Fixed missing shifts in utl_field_get and utl_field_set
*   04-Oct-94  DMW   - Added "{" to "for" loop in utl_in_range_2.
*                    - Substituted spaces for tabs.
*                    - Changed all instances of ULONG to UINT
*   09-Oct-94  DMW   - Fix utl_encode_2, shift '1', not 2.
*   22-Mar-95  TLR   - Added utl_buff_zero and utl_buff_copy.
*   23-Mar-95  DMW   - Changed names of utl_buff_zero and utl_buff_copy to 
*                      utl_mem_zero and utl_mem_copy.                 
*
* Description:
*   None
*/

#include "g_types.h"
#include "codes.h"
#include "utl.h"

/*++
* ================
* = utl_encode_2 =
* ================
*
* Description:
*   The encoding is done by starting with the power of 2 value that corresponds
*   to the encoded value of 0, and comparing it to the value to encode.  If it
*   does not match, the encoding is incremented and the value to compare is
*   doubled.
*
*   To guard against an infinite loop, the search uses a "<" operator so that
*   the search is guaranteed to complete.
*
* Issues: 
*   None
--*/
UINT32 utl_encode_2 (UINT32 encodes_as_0, UINT32 value_to_encode)
   {
   UINT32 encoding, tmp;
   
   /* Start with an encoding of 0, search until a match is found. */
   
   encoding = 0;
   tmp      = encodes_as_0;

   while (tmp < value_to_encode)
      {
      encoding += 1;
      tmp       = tmp << 1;
      }
   
   return (encoding);
   }

/*++
* =================
* = utl_field_get =
* =================
*
* Description:
*   The field is described by a bit mask only.  The bit position of the field 
*   is determined by finding the least significant set bit.  The bit position
*   is needed to right-justify specified field value to the correct postion for 
*   the "and" operation.
*
*   A different implementation of this routine may have not calculated the bit
*   position of the mask, but instead shifted the new field value as the mask was
*   being shifted.  However, this is more convoluted and the savings are minimal.
*
* Issues: 
*   None
--*/
UINT32 utl_field_get (UINT32 field_value, UINT32 field_mask)
   {
   int bit_position;
   
   /* Return 0 if the field mask 0, otherwise subsequent code will */
   /*   get caught in an infinite loop.                            */
   
   if (field_mask == 0) return (0);
   
   /* Next calculate the bit position of the field. */
   
   bit_position = 0;
   
   while ((field_mask & 1) == 0)
      {
      field_mask    = (field_mask >> 1);
      bit_position += 1;
      }
      
   /* Return the value extracted from the field. */
   
   return ((field_value >> (bit_position)) & field_mask);
   }

/*++
* =================
* = utl_field_set =
* =================
*
* Description:
*   To set a field to a certain value, first the field must be cleared because
*   the field is set by an "or" operation.  If the field were not clear, then
*   the result would be incorrect.
*
*   The field is described by a bit mask only.  The bit position of the field 
*   is determined by finding the least significant set bit.  The bit position
*   is needed to shift the specified field value to the correct postion for 
*   the "or" operation.
*
*   A different implementation of this routine may have not calculated the bit
*   position of the mask, but instead shifted the new field value as the mask was
*   being shifted.  However, this is more convoluted the and savings are minimal.
*
* Testing:
*   To adequately test this routine, the following cases should be evaluated:
*     - *p_value = DON'T CARE, field_mask = DON'T CARE, field_value = 0x00000000
*     - *p_value = 0x00000000, field_mask = 0x0000000F, field_value = 0x0000000F
*     - *p_value = 0x0000000F, field_mask = 0x0000000F, field_value = 0x00000000
*     - *p_value = 0x00000000, field_mask = 0xF0000000, field_value = 0xF0000000
*
* Issues: 
*   None
--*/
void utl_field_set (UINT32 *p_value, UINT32 field_mask, UINT32 field_value)
   {
   int bit_position;
   
   /* Return if the field mask is the value 0, otherwise subsequent code will */
   /*   get caught in an infinite loop.                                       */
   
   if (field_mask == 0) return;
   
   /* First clear out the field using the field mask. */
   
   *p_value &= ~field_mask;
    
   /* Next calculate the bit position of the field. */
   
   bit_position = 0;
   
   while ((field_mask & 1) == 0)
      {
      field_mask    = (field_mask >> 1);
      bit_position += 1;
      }
   
   /* Now update the field. */
   
   *p_value |= (field_value << bit_position);
   }


/*++
* =====================
* = utl_field_set_def =
* =====================
*
* Description:
*   Rather than passing the bit position of the field, the mask (which has a 
*   set '1' for each bit in the field) is used to calculate the bit position
*   of the field.  This is accomplished by counting the bits from the LSB until
*   the first set bit in the mask.
*
* Testing:
*   - p_value = 0x00000000, field_mask = 0x0000000F, field_value = DEFAULT
*   - p_value = 0x00000000, field_mask = 0x0000000F, field_value = 0x00000001
*
* Issues:
*   None
--*/


void utl_field_set_def (UINT32 *p_value, UINT32 field_mask, UINT32 field_value)
   {
   if (field_value != DEFAULT)
      utl_field_set (p_value, field_mask, field_value);
   }

/*++
* ================
* = utl_field_ok =
* ================
*
* Description:
*   In order for the value to be considered acceptable, it must fit into the
*   specified field.
*
* Testing:
*   To adequately test this routine, the following cases should be evaluated:
*     - value = 0,          mask = 0x00000000 
*     - value = 0x0000000F, mask = 0x0000000F
*     - value = 0x0000000F, mask = 0xF0000000
*     - value = 0x0000001F, mask = 0xF0000000
*
* Issues: 
*   None
--*/
BOOLEAN utl_field_ok (UINT32 value, UINT32 mask)
   {
   /* Return FALSE if the field has a width of 0. */
   
   if (mask == 0) return (FALSE);

   /* Shift the mask right until it is right-justified. */
   
   while ((mask & 1) == 0)
      mask = (mask >> 1);
   
   /* Check if the value fits in the field. */
   
   if ((value & ~mask) != 0)
      return (FALSE);
   else
      return (TRUE);
   }

/*++
* ====================
* = utl_field_ok_def =
* ====================
*
* Description:
*   In order for the value to be considered acceptable, it must fit into the
*   specified field or have the value DEFAULT.  The value of DEFAULT works on
*   fields which are 0 bytes wide.
*
* Testing:
*   To adequately test this routine, the following cases should be evaluated:
*     - value = 8,          mask = 0x0000000F 
*     - value = DEFAULT,    mask = 0xFFFFFFFF
*
* Issues: 
*   None
--*/
BOOLEAN utl_field_ok_def (UINT32 value, UINT32 mask)
   {
   /* Return TRUE if the value is DEFAULT. */
   
   if (value == DEFAULT)
      return (TRUE);
   else
     return (utl_field_ok (value, mask));
   }

/*++
* ================
* = utl_in_range =
* ================
*
* Description:
*   In order for the value to be considered acceptable, it must fall into the
*   specified range or have the value DEFAULT.
*
* Testing:
*   To adequately test this routine, the following cases should be evaluated:
*     - value < min
*     - value > min
*     - value = min
*     - value = max
*     - value = 0 (min = 0)
*     - value = 1 (min = 0)
*     - value = 0xFFFFFFFF (max = 0xFFFFFFFF)
*     - value = 8 (min = 0, max = 16)
*     - value = 3 (min = 0, max = 16)
*
* Issues: 
*   Add a test case for value > max.  Also, what about min > max?
--*/
BOOLEAN utl_in_range (UINT32 value, UINT32 min, UINT32 max)
   {
   /* Return failure if the value is out of range. */
   
   if ((value < min) || (value > max))
      return (FALSE);
   else    
      return (TRUE);
   }

/*++
* ====================
* = utl_in_range_def =
* ====================
*
* Description:
*   In order for the value to be considered acceptable, it must fall into the
*   specified range or have the value DEFAULT.
*
* Testing:
*   To adequately test this routine, the following cases should be evaluated:
*     - value = DEFAULT.
*
* Issues: 
*   None
--*/
BOOLEAN utl_in_range_def (UINT32 value, UINT32 min, UINT32 max)
   {
   /* Return success if the value is default. */
   
   if (value == DEFAULT)
      return (TRUE);
   else   
      return (utl_in_range (value, min, max));
   }

/*++
* ==================
* = utl_in_range_2 =
* ==================
*
* Description:
*   In order for the value to be considered acceptable, it must fall into the
*   specified range and have fewer than 2 bits set.
*
* Testing:
*   To adequately test this routine, the following cases should be evaluated:
*     - value < min
*     - value > min
*     - value = min
*     - value = max
*     - value = 0 (min = 0)
*     - value = 1 (min = 0)
*     - value = 0xFFFFFFFF (max = 0xFFFFFFFF)
*     - value = 8 (min = 0, max = 16)
*     - value = 3 (min = 0, max = 16)
*
* Issues: 
*   - Is the second case correct?  Should it be value > max?
*   - Need a test case of 0x80000000 that succeeds.
--*/
BOOLEAN utl_in_range_2 (UINT32 value, UINT32 min, UINT32 max)
   {
   int   i, set_bit_cnt;
   
   /* Return failure if the value is out of range. */
   
   if ((value < min) || (value > max))
      return (FALSE);

   /* Count the number of bits in the value. */

   set_bit_cnt = 0;
   
   for (i = 0; i < 32; i++)
      {
      if (value & (1 << i) != 0)
         set_bit_cnt += 1;
      }
   
   /* Failure if more than 1 bit is set, because it can't be a power of 2 */
   
   if (set_bit_cnt > 1)
      return (FALSE);
   else
      return (TRUE);
   } 

/*++
* ======================
* = utl_in_range_2_def =
* ======================
*
* Description:
*   In order for the value to be considered acceptable, it must fall into the
*   specified range and have fewer than 2 bits set.
*
* Testing:
*   To adequately test this routine, the following cases should be evaluated:
*     - value = DEFAULT
*
* Issues: 
*   None
--*/
BOOLEAN utl_in_range_2_def (UINT32 value, UINT32 min, UINT32 max)
   {
   if (value == DEFAULT)
      return (TRUE);
   else
      return (utl_in_range_2 (value, min, max));
   }
   
/*++
* ================
* = utl_mem_zero =
* ================
*
* Description:
*   This routine is not intended to be blazing fast, but instead is designed
*   to be simple.
*
* Issues: 
*   None
--*/
void utl_mem_zero (void  *p_buf,
                   int    size)
   {
   int    i;
   UINT8 *p_array;

   p_array = (UINT8 *) p_buf;

   for (i = 0; i < size; i++)
      p_array[i] = 0;
   }

/*++
* ================
* = utl_mem_copy =
* ================
*
* Description:
*   Copies the bytes from the source buffer to the destination buffer.  This
*   routine is not intended to be blazing fast, but instead is designed to be
*   simple.
*
* Issues: 
*   None
--*/
void utl_mem_copy (void  *p_from,
                   void  *p_to,
                   int    size)
   {
   int    i;
   UINT8 *p_src;
   UINT8 *p_dst;

   p_src = (UINT8 *) p_from;
   p_dst = (UINT8 *) p_to;

   for (i = 0; i < size; i++)
      p_dst[i] = p_src[i];
   }

